#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "cfg.h"

struct SSettings Settings;


char bootDevice[10] = "SD:";

u8 ocarinaChoice = 0;
u8 videoChoice = 0;
u8 languageChoice = 0;
u8 viChoice = 0;
u8 iosChoice = 0;
u8 parentalcontrolChoice = 0;
u8 fix002 = 0;
u8 reloadblock = 0;
u8 countrystrings = 0;
u8 alternatedol = 0;
u32 alternatedoloffset = 0;
char alternatedname[40];

#define TITLE_MAX 65

struct ID_Title
{
	u8 id[5];
	char title[TITLE_MAX];
};

struct ID_Control
{
	u8 id[5];
	u8 block;
};
// renamed titles
int num_title = 0; //number of titles
struct ID_Title *cfg_title = NULL;

int num_control = 0;
struct ID_Control *cfg_control = NULL;

#define MAX_SAVED_GAMES 1000
#define MAX_SAVED_GAME_NUM 1000
int num_saved_games = 0;
struct Game_CFG cfg_game[MAX_SAVED_GAMES];


/* For Mapping */

static char *cfg_name, *cfg_val;

struct TextMap
{
	char *name;
	int id;
};

struct TextMap map_video[] =
{
	{ "system", CFG_VIDEO_SYS },
	{ "game",   CFG_VIDEO_GAME },
	{ "patch",  CFG_VIDEO_PATCH },
	{ "pal50", CFG_VIDEO_PAL50 },
	{ "pal60", CFG_VIDEO_PAL60 },
	{ "ntsc", CFG_VIDEO_NTSC },
	{ NULL, -1 }
};

struct TextMap map_language[] =
{
	{ "console",   CFG_LANG_CONSOLE },
	{ "japanese",  CFG_LANG_JAPANESE },
	{ "english",   CFG_LANG_ENGLISH },
	{ "german",    CFG_LANG_GERMAN },
	{ "french",    CFG_LANG_FRENCH },
	{ "spanish",   CFG_LANG_SPANISH },
	{ "italian",   CFG_LANG_ITALIAN },
	{ "dutch",     CFG_LANG_DUTCH },
	{ "s.chinese", CFG_LANG_S_CHINESE },
	{ "t.chinese", CFG_LANG_T_CHINESE },
	{ "korean",    CFG_LANG_KOREAN },
	{ NULL, -1 }
};


struct TextMap map_alignment[] =
{
	{ "left",   CFG_ALIGN_LEFT },
	{ "right",  CFG_ALIGN_RIGHT },
	{ "center",   CFG_ALIGN_CENTRE },
	{ "top",    CFG_ALIGN_TOP },
	{ "bottom",    CFG_ALIGN_BOTTOM },
	{ "middle",   CFG_ALIGN_MIDDLE },
	{ NULL, -1 }
};
char* strcopy(char *dest, char *src, int size)
{
	strncpy(dest,src,size);
	dest[size-1] = 0;
	return dest;
}

int map_get_id(struct TextMap *map, char *name)
{
	int i;
	for (i=0; map[i].name != NULL; i++)	{
		if (strcmp(name, map[i].name) == 0) return map[i].id;
	}
	return -1;
}

char* map_get_name(struct TextMap *map, short id)
{
	int i;
	for (i=0; map[i].name != NULL; i++)	{
		if (id == map[i].id) return map[i].name;
	}
	return NULL;
}

bool map_auto(char *name, char *name2, char *val, struct TextMap *map, short *var)
{
	if (strcmp(name, name2) != 0) return false;
 	int id = map_get_id(map, val);
	if (id == -1) {
		//printf("MAP FAIL: %s=%s : %d\n", name, val, id); sleep(1);
		return false;
	}
	*var = id;
	//printf("MAP AUTO: %s=%s : %d\n", name, val, id); sleep(1);
	return true;
}

bool cfg_map_auto(char *name, struct TextMap *map, short *var)
{
	return map_auto(name, cfg_name, cfg_val, map, var);
}

bool cfg_map(char *name, char *val, short *var, short id)
{
	if (strcmp(name, cfg_name)==0 && strcmpi(val, cfg_val)==0)
	{
		*var = id;
		return true;
	}
	return false;
}

bool cfg_bool(char *name, short *var)
{
	return (cfg_map(name, "0", var, 0) || cfg_map(name, "1", var, 1));
}

void cfg_int(char *name, short *var, int count)
{
	char tmp[5];
	short i;

	if (count > 10) //avoid overflow
		return;

	for (i = 0; i < count; i++)
	{
		sprintf(tmp, "%d", i);
		cfg_map(name, tmp, var, i);
	}
}

/* Mapping */

//static char bg_path[100];

void CFG_Default(int widescreen) // -1 = non forced Mode
{
	if (widescreen == -1)
	{
		snprintf(Settings.covers_path, sizeof(Settings.covers_path), "%s/images/", bootDevice); //default image path
		snprintf(Settings.disc_path, sizeof(Settings.disc_path), "%s/images/disc/", bootDevice);
		snprintf(Settings.titlestxt_path, sizeof(Settings.titlestxt_path), "%s/config/", bootDevice);//default path for disc images
		snprintf(Settings.language_path, sizeof(Settings.language_path), "notset");
		snprintf(Settings.languagefiles_path, sizeof(Settings.languagefiles_path), "%s/config/language/", bootDevice);
		snprintf(Settings.oggload_path, sizeof(Settings.oggload_path), "%s/config/backgroundmusic/", bootDevice);
		snprintf(Settings.update_path, sizeof(Settings.update_path), "%s/apps/usbloader_gx/", bootDevice);
		snprintf(Settings.Cheatcodespath, sizeof(Settings.Cheatcodespath), "%s/codes/", bootDevice);
		snprintf(Settings.TxtCheatcodespath, sizeof(Settings.TxtCheatcodespath), "%s/txtcodes/", bootDevice);
		snprintf(Settings.dolpath, sizeof(Settings.dolpath), "%s/", bootDevice);
		sprintf(Settings.ogg_path, "notset");
	}
}

void Global_Default(void)
{
	Settings.video = discdefault;
	Settings.vpatch = off;
	Settings.language = ConsoleLangDefault;
	Settings.ocarina = off;
	Settings.parentalcontrol = 0;
	Settings.cios = ios249;
	Settings.wiilight = 1;
	Settings.patchcountrystrings = 0;
	Settings.error002 = 0;
}


char *cfg_get_title(u8 *id)
{
	int i;
	for (i=0; i<num_title; i++) {
		if (memcmp(id, cfg_title[i].id, 4) == 0) {
			return cfg_title[i].title;
		}
	}
	return NULL;
}

char *get_title(struct discHdr *header)
{
	char *title = cfg_get_title(header->id);
	if (title) return title;
	return header->title;
}

void title_set(char *id, char *title)
{
	char *idt = cfg_get_title((u8*)id);
	if (idt) {
		// replace
		strcopy(idt, title, TITLE_MAX);
	} else {
		cfg_title = realloc(cfg_title, (num_title+1) * sizeof(struct ID_Title));
		if (!cfg_title) {
			// error
			num_title = 0;
			return;
		}
		// add
		memcpy(cfg_title[num_title].id, id, 4);
		cfg_title[num_title].id[4] = 0;
		strcopy(cfg_title[num_title].title, title, TITLE_MAX);
		num_title++;
	}
}

u8 cfg_get_block(u8 *id)
{
	int i;
	for (i=0; i<num_control; i++) {
		if (memcmp(id, cfg_control[i].id, 4) == 0) {
			return cfg_control[i].block;
		}
	}
	return 0;
}

u8 get_block(struct discHdr *header)
{
	return cfg_get_block(header->id);
}

// trim leading and trailing whitespace
// copy at max n or at max size-1
char* trim_n_copy(char *dest, char *src, int n, int size)
{
	int len;
	// trim leading white space
	while (isspace(*src)) { src++; n--; }
	len = strlen(src);
	if (len > n) len = n;
	// trim trailing white space
	while (len > 0 && isspace(src[len-1])) len--;
	if (len >= size) len = size-1;
	strncpy(dest, src, len);
	dest[len] = 0;
	//printf("trim_copy: '%s' %d\n", dest, len); //sleep(1);
	return dest;
}

char* trimcopy(char *dest, char *src, int size)
{
	int len;
	while (*src == ' ') src++;
	len = strlen(src);
	// trim trailing " \r\n"
	while (len > 0 && strchr(" \r\n", src[len-1])) len--;
	if (len >= size) len = size-1;
	strncpy(dest, src, len);
	dest[len] = 0;
	return dest;
}

void path_set(char *name, char *val)
{
	cfg_name = name;
	cfg_val = val;

	// if these are defined in txt file, use them.  otherwise use defaults

	if (strcmp(name, "cover_path") == 0) {
		strcopy(Settings.covers_path, val, sizeof(Settings.covers_path));
		return;
	}

	if (strcmp(name, "disc_path") == 0) {
		strcopy(Settings.disc_path, val, sizeof(Settings.disc_path));
		return;
	}
    if (strcmp(name, "titlestxt_path") == 0) {
		strcopy(Settings.titlestxt_path, val, sizeof(Settings.titlestxt_path));
		return;
	}
	if (strcmp(name, "language_path") == 0) {
		strcopy(Settings.language_path, val, sizeof(Settings.language_path));
		return;
	}
	if (strcmp(name, "languagefiles_path") == 0) {
		strcopy(Settings.languagefiles_path, val, sizeof(Settings.languagefiles_path));
		return;
	}
	if (strcmp(name, "update_path") == 0) {
		strcopy(Settings.update_path, val, sizeof(Settings.update_path));
		return;
	}
	if (strcmp(name, "Cheatcodespath") == 0) {
		strcopy(Settings.Cheatcodespath, val, sizeof(Settings.Cheatcodespath));
		return;
	}
	if (strcmp(name, "TxtCheatcodespath") == 0) {
		strcopy(Settings.TxtCheatcodespath, val, sizeof(Settings.TxtCheatcodespath));
		return;
	}
	if (strcmp(name, "oggload_path") == 0) {
		strcopy(Settings.oggload_path, val, sizeof(Settings.oggload_path));
		return;
	}
    if (strcmp(name, "dolpath") == 0) {
		strcopy(Settings.dolpath, val, sizeof(Settings.dolpath));
		return;
	}
    if (strcmp(name, "ogg_path") == 0) {
		strcopy(Settings.ogg_path, val, sizeof(Settings.ogg_path));
		return;
	}

	return;

}

void global_cfg_set(char *name, char *val)
{
	cfg_name = name;
	cfg_val = val;

	if (strcmp(name, "video") == 0) {
	        int i;
		if (sscanf(val, "%d", &i) == 1) {
			Settings.video = i;
		}
		return;
	}
	else if (strcmp(name, "vpatch") == 0) {
		int i;
		if (sscanf(val, "%d", &i) == 1) {
			Settings.vpatch =i;
		}
		return;
	}

	else if (strcmp(name, "language") == 0) {
		int i;
		if (sscanf(val, "%d", &i) == 1) {
			Settings.language = i;
		}
		return;
	}
	else if (strcmp(name, "ocarina") == 0) {
		int i;
		if (sscanf(val, "%d", &i) == 1) {
			Settings.ocarina = i;
		}
		return;
	}
	else if (strcmp(name, "parentalcontrol") == 0) {
		int i;
		if (sscanf(val, "%d", &i) == 1) {
			Settings.parentalcontrol = i;
		}
		return;
	}
	else if (strcmp(name, "cios") == 0) {
		int i;
		if (sscanf(val, "%d", &i) == 1) {
            Settings.cios = i;
			}
		return;
	}
	else if (strcmp(name, "wiilight") == 0) {
		int i;
		if (sscanf(val, "%d", &i) == 1) {
            Settings.wiilight = i;
			}
		return;
	}
	else if (strcmp(name, "patchcountrystrings") == 0) {
		int i;
		if (sscanf(val, "%d", &i) == 1) {
            Settings.patchcountrystrings = i;
			}
		return;
	}
	else if (strcmp(name, "error002") == 0) {
		int i;
		if (sscanf(val, "%d", &i) == 1) {
            Settings.error002 = i;
			}
		return;
	}

	return;
}

// split line to part1 delimiter part2
bool trimsplit(char *line, char *part1, char *part2, char delim, int size)
{
	char *eq = strchr(line, delim);
	if (!eq) return false;
	trim_n_copy(part1, line, eq-line, size);
	trimcopy(part2, eq+1, size);
	return true;
}
void cfg_parseline(char *line, void (*set_func)(char*, char*))
{
	// split name = value
	char tmp[200], name[200], val[200];
	strcopy(tmp, line, sizeof(tmp));
	char *eq = strchr(tmp, '=');
	if (!eq) return;
	*eq = 0;
	trimcopy(name, tmp, sizeof(name));
	trimcopy(val, eq+1, sizeof(val));
	//printf("CFG: %s = %s\n", name, val);
	set_func(name, val);
}

void cfg_parsetitleline(char *line, void (*set_func)(char*, char*, u8))
{
	// split name = value
	char tmp[200], name[200], val[200];
	int block = 0;
	strcopy(tmp, line, sizeof(tmp));
	char *eq = strchr(tmp, '=');
	if (!eq) return;
	*eq = 0;
	trimcopy(name, tmp, sizeof(name));

	char *blockpos = strrchr(eq+1, '=');

	if (!blockpos)
		trimcopy(val, eq+1, sizeof(val));

	else
	{
		*blockpos = 0;
		trimcopy(val, eq+1, sizeof(val));
		if (sscanf(blockpos+1, "%d", &block) != 1)
		{
			block = 0;
		}
	}
	set_func(name, val, block);
}

bool cfg_parsefile(char *fname, void (*set_func)(char*, char*))
{
	FILE *f;
	char line[200];

	//printf("opening(%s)\n", fname);
	f = fopen(fname, "r");
	if (!f) {
		//printf("error opening(%s)\n", fname);
		return false;
	}
	while (fgets(line, sizeof(line), f)) {
		// lines starting with # are comments
		if (line[0] == '#') continue;
		cfg_parseline(line, set_func);
	}
	fclose(f);
	return true;
}

bool cfg_parsetitlefile(char *fname, void (*set_func)(char*, char*, u8))
{
	FILE *f;
	char line[200];

	//printf("opening(%s)\n", fname);
	f = fopen(fname, "r");
	if (!f) {
		//printf("error opening(%s)\n", fname);
		return false;
	}

	while (fgets(line, sizeof(line), f)) {
		// lines starting with # are comments
		if (line[0] == '#') continue;
		cfg_parsetitleline(line, set_func);
	}
	fclose(f);
	return true;
}

// PER-GAME SETTINGS
// return existing or new
struct Game_CFG* cfg_get_game(u8 *id)
{
	struct Game_CFG *game = CFG_get_game_opt(id);
	if (game) return game;
	if (num_saved_games >= MAX_SAVED_GAMES) return NULL;
	game = &cfg_game[num_saved_games];
	num_saved_games++;
	return game;
}

// current options to game
void cfg_set_game_opt(struct Game_CFG *game, u8 *id)
{
	strncpy((char*)game->id, (char*)id, 6);
	game->id[6] = 0;
	game->video = videoChoice;
	game->language = languageChoice;
	game->ocarina = ocarinaChoice;
	game->vipatch = viChoice;
	game->ios = iosChoice;
	game->parentalcontrol = parentalcontrolChoice;
	game->errorfix002 = fix002;
	game->iosreloadblock = reloadblock;
	game->patchcountrystrings = countrystrings;
	game->loadalternatedol = alternatedol;
	if (game->loadalternatedol == 0) {
		alternatedoloffset = 0;
		strcpy(alternatedname, "");
	}
	game->alternatedolstart = alternatedoloffset;
	strcopy(game->alternatedolname, alternatedname,sizeof(game->alternatedolname));
}

void game_set(char *name, char *val)
{
	// sample line:
	// game:RTNP41 = video:game; language:english; ocarina:0;
	// game:RYWP01 = video:patch; language:console; ocarina:1;
	//printf("GAME: '%s=%s'\n", name, val);
	u8 id[8];
	struct Game_CFG *game;
	if (strncmp(name, "game:", 5) != 0) return;
	trimcopy((char*)id, name+5, sizeof(id));
	game = cfg_get_game(id);
	// set id and current options as default
	cfg_set_game_opt(game, id);
	//printf("GAME(%s) '%s'\n", id, val); sleep(1);

	// parse val
	// first split options by ;
	char opt[200], *p, *np;
	p = val;

	while(p) {
		np = strchr(p, ';');
		if (np) trim_n_copy(opt, p, np-p, sizeof(opt));
		else trimcopy(opt, p, sizeof(opt));
		//printf("GAME(%s) (%s)\n", id, opt); sleep(1);
		// parse opt 'language:english'
		char opt_name[200], opt_val[200];
		if (trimsplit(opt, opt_name, opt_val, ':', sizeof(opt_name))){
			//printf("GAME(%s) (%s=%s)\n", id, opt_name, opt_val); sleep(1);
			short opt_v, opt_l, opt_c;
			if (map_auto("video", opt_name, opt_val, map_video, &opt_v)) {
				// valid option, assign
				game->video = opt_v;
			}
			if (map_auto("language", opt_name, opt_val, map_language, &opt_l)) {
				// valid option, assign
				game->language = opt_l;
			}
			if (strcmp("ocarina", opt_name) == 0) {
				if (sscanf(opt_val, "%hd", &opt_c) == 1) {
					game->ocarina = opt_c;
				}
			}
			if (strcmp("vipatch", opt_name) == 0) {
				if (sscanf(opt_val, "%hd", &opt_c) == 1) {
					game->vipatch = opt_c;
				}
			}
			if (strcmp("ios", opt_name) == 0) {
				if (sscanf(opt_val, "%hd", &opt_c) == 1) {
					game->ios = opt_c;
				}
			}
			if (strcmp("pctrl", opt_name) == 0) {
				if (sscanf(opt_val, "%hd", &opt_c) == 1) {
					game->parentalcontrol = opt_c;
				}
			}
			if (strcmp("errorfix002", opt_name) == 0) {
				if (sscanf(opt_val, "%hd", &opt_c) == 1) {
					game->errorfix002 = opt_c;
				}
			}
			if (strcmp("iosreloadblock", opt_name) == 0) {
				if (sscanf(opt_val, "%hd", &opt_c) == 1) {
					game->iosreloadblock = opt_c;
				}
			}
			if (strcmp("patchcountrystrings", opt_name) == 0) {
				if (sscanf(opt_val, "%hd", &opt_c) == 1) {
					game->patchcountrystrings = opt_c;
				}
			}
			if (strcmp("loadalternatedol", opt_name) == 0) {
				if (sscanf(opt_val, "%hd", &opt_c) == 1) {
					game->loadalternatedol = opt_c;
				}
			}
			if (strcmp("alternatedolstart", opt_name) == 0) {
				if (sscanf(opt_val, "%hd", &opt_c) == 1) {
					game->alternatedolstart = opt_c;
				}
			}
			if (strcmp("alternatedolname", opt_name) == 0) {
				strcopy(game->alternatedolname, opt_val, sizeof(game->alternatedolname));
			}
		}
		// next opt
		if (np) p = np + 1; else p = NULL;
	}
}

void parental_set(char *name, char *val)
{
	// sample line:
	// game:RTNP41 = video:game; language:english; ocarina:0;
	// game:RYWP01 = video:patch; language:console; ocarina:1;
	//printf("GAME: '%s=%s'\n", name, val);
	u8 id[8];

	if (strncmp(name, "game:", 5) != 0) return;
	trimcopy((char*)id, name+5, sizeof(id));

	// parse val
	// first split options by ;
	char opt[200], *p, *np;
	p = val;

	while(p) {
		np = strchr(p, ';');
		if (np) trim_n_copy(opt, p, np-p, sizeof(opt));
		else trimcopy(opt, p, sizeof(opt));
		//printf("GAME(%s) (%s)\n", id, opt); sleep(1);
		// parse opt 'language:english'
		char opt_name[200], opt_val[200];
		if (trimsplit(opt, opt_name, opt_val, ':', sizeof(opt_name))){
			//printf("GAME(%s) (%s=%s)\n", id, opt_name, opt_val); sleep(1);
			short opt_c;

			if (strcmp("pctrl", opt_name) == 0) {
				if (sscanf(opt_val, "%hd", &opt_c) == 1) {
					cfg_control = realloc(cfg_control, (num_control+1) * sizeof(struct ID_Control));
					if (!cfg_control) {
					// error
						num_control = 0;
						return;
					}
					// add
					memcpy(cfg_control[num_control].id, id, 4);
					cfg_control[num_control].id[4] = 0;
					cfg_control[num_control].block = opt_c;
					num_control++;
				}
			}

		}
		// next opt
		if (np) p = np + 1; else p = NULL;
	}
}

bool cfg_load_games()
{
	char GXGameSettings_cfg[32];
	sprintf(GXGameSettings_cfg, "%s/config/GXGameSettings.cfg", bootDevice);
	return cfg_parsefile(GXGameSettings_cfg, &game_set);
}

bool cfg_load_global()
{
	char GXGlobal_cfg[26];
	sprintf(GXGlobal_cfg, "%s/config/GXGlobal.cfg", bootDevice);
	//Default values defined by dev team
	Settings.video = discdefault;
	Settings.vpatch = off;
	Settings.language = ConsoleLangDefault;
	Settings.ocarina = off;
	return cfg_parsefile(GXGlobal_cfg, &global_cfg_set);
}

struct Game_CFG* CFG_get_game_opt(u8 *id)
{
	int i;
	for (i=0; i<num_saved_games; i++) {
		if (memcmp(id, cfg_game[i].id, 6) == 0) {
			return &cfg_game[i];
		}
	}
	return NULL;
}

void CFG_Load(void)
{
	char pathname[200];

	CFG_Default(-1); // set defaults non forced

	snprintf(pathname, sizeof(pathname), "%s/config/GXGlobal.cfg", bootDevice);

	cfg_parsefile(pathname, &path_set); //then set config and layout options

	snprintf(pathname, sizeof(pathname), "%s/config/GXGameSettings.cfg", bootDevice);
	cfg_parsefile(pathname, &parental_set);

	// load per-game settings
	cfg_load_games();

	Global_Default(); //global default depends on theme information
	CFG_LoadGlobal();

}

void CFG_LoadGlobal(void)
{
	char GXGlobal_cfg[26];
	sprintf(GXGlobal_cfg, "%s/config/GXGlobal.cfg", bootDevice);
	cfg_parsefile(GXGlobal_cfg, &global_cfg_set);
}

void CFG_Cleanup(void)
{
	if (cfg_title)
	{
		free(cfg_title);
		cfg_title = NULL;
	}
}
